@@ -32,6 +32,12 @@ public class Document
 	public static final String META_ENCRYPTION = "encryption";
 	public static final String META_INFO_AUTHOR = "info:Author";
 	public static final String META_INFO_TITLE = "info:Title";
+	public static final String META_INFO_SUBJECT = "info:Subject";
+	public static final String META_INFO_KEYWORDS = "info:Keywords";
+	public static final String META_INFO_CREATOR = "info:Creator";
+	public static final String META_INFO_PRODUCER = "info:Producer";
+	public static final String META_INFO_CREATIONDATE = "info:CreationDate";
+	public static final String META_INFO_MODIFICATIONDATE = "info:ModDate";
 
 	protected long pointer;
 
@@ -80,6 +86,7 @@ public class Document
 
 	public static native boolean recognize(String magic);
 
+	public native boolean supportsAccelerator();
 	public native void saveAccelerator(String filename);
 	public native void outputAccelerator(SeekableOutputStream stream);
 
@@ -185,7 +192,7 @@ public class Document
 		return -1;
 	}
 
-	public native Quad[] search(int chapter, int page, String needle);
+	public native Quad[][] search(int chapter, int page, String needle);
 
 	public native Location resolveLink(String uri);
 	public Location resolveLink(Outline link) {
@@ -195,8 +202,21 @@ public class Document
 		return resolveLink(link.uri);
 	}
 
+	public native LinkDestination resolveLinkDestination(String uri);
+	public LinkDestination resolveLinkDestination(OutlineIterator.OutlineItem item) {
+		return resolveLinkDestination(item.uri);
+	}
+	public LinkDestination resolveLinkDestination(Outline link) {
+		return resolveLinkDestination(link.uri);
+	}
+	public LinkDestination resolveLinkDestination(Link link) {
+		return resolveLinkDestination(link.uri);
+	}
+
 	public native Outline[] loadOutline();
+	public native OutlineIterator outlineIterator();
 	public native String getMetaData(String key);
+	public native void setMetaData(String key, String value);
 	public native boolean isReflowable();
 	public native void layout(float width, float height, float em);
 
@@ -215,8 +235,9 @@ public class Document
 
 	public native boolean isUnencryptedPDF();
 
+	public native String formatLinkURI(LinkDestination dest);
+
 	public boolean isPDF() {
 		return false;
 	}
-
 }
