@@ -42,9 +42,6 @@
 #include FT_TRUETYPE_TABLES_H
 #include FT_TRUETYPE_TAGS_H
 
-#define MAX_BBOX_TABLE_SIZE 4096
-#define MAX_ADVANCE_CACHE 4096
-
 #ifndef FT_SFNT_OS2
 #define FT_SFNT_OS2 ft_sfnt_os2
 #endif
@@ -98,7 +95,6 @@ static fz_font *
 fz_new_font(fz_context *ctx, const char *name, int use_glyph_bbox, int glyph_count)
 {
 	fz_font *font;
-	int i;
 
 	font = fz_malloc_struct(ctx, fz_font);
 	font->refs = 1;
@@ -130,22 +126,8 @@ fz_new_font(fz_context *ctx, const char *name, int use_glyph_bbox, int glyph_cou
 
 	font->glyph_count = glyph_count;
 
-	if (use_glyph_bbox && glyph_count <= MAX_BBOX_TABLE_SIZE)
-	{
-		fz_try(ctx)
-			font->bbox_table = Memento_label(fz_malloc_array(ctx, glyph_count, fz_rect), "font_bbox_table");
-		fz_catch(ctx)
-		{
-			fz_free(ctx, font);
-			fz_rethrow(ctx);
-		}
-		for (i = 0; i < glyph_count; i++)
-			font->bbox_table[i] = fz_infinite_rect;
-	}
-	else
-	{
-		font->bbox_table = NULL;
-	}
+	font->bbox_table = NULL;
+	font->use_glyph_bbox = use_glyph_bbox;
 
 	font->width_count = 0;
 	font->width_table = NULL;
@@ -239,9 +221,21 @@ fz_drop_font(fz_context *ctx, fz_font *font)
 		fz_free(ctx, font->encoding_cache[i]);
 
 	fz_drop_buffer(ctx, font->buffer);
-	fz_free(ctx, font->bbox_table);
+	if (font->bbox_table)
+	{
+		int n = (font->glyph_count+255)/256;
+		for (i = 0; i < n; i++)
+			fz_free(ctx, font->bbox_table[i]);
+		fz_free(ctx, font->bbox_table);
+	}
 	fz_free(ctx, font->width_table);
-	fz_free(ctx, font->advance_cache);
+	if (font->advance_cache)
+	{
+		int n = (font->glyph_count+255)/256;
+		for (i = 0; i < n; i++)
+			fz_free(ctx, font->advance_cache[i]);
+		fz_free(ctx, font->advance_cache);
+	}
 	if (font->shaper_data.destroy && font->shaper_data.shaper_handle)
 	{
 		font->shaper_data.destroy(ctx, font->shaper_data.shaper_handle);
@@ -1200,6 +1194,30 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix tr
 	return result;
 }
 
+static fz_rect *
+get_gid_bbox(fz_context *ctx, fz_font *font, int gid)
+{
+	int i;
+
+	if (gid < 0 || gid >= font->glyph_count || !font->use_glyph_bbox)
+		return NULL;
+
+	if (font->bbox_table == NULL) {
+		i = (font->glyph_count + 255)/256;
+		font->bbox_table = Memento_label(fz_malloc_array(ctx, i, fz_rect *), "bbox_table(top)");
+		memset(font->bbox_table, 0, sizeof(fz_rect *) * i);
+	}
+
+	if (font->bbox_table[gid>>8] == NULL) {
+		font->bbox_table[gid>>8] = Memento_label(fz_malloc_array(ctx, 256, fz_rect), "bbox_table");
+		for (i = 0; i < 256; i++) {
+			font->bbox_table[gid>>8][i] = fz_infinite_rect;
+		}
+	}
+
+	return &font->bbox_table[gid>>8][gid & 255];
+}
+
 static fz_rect *
 fz_bound_ft_glyph(fz_context *ctx, fz_font *font, int gid)
 {
@@ -1208,7 +1226,7 @@ fz_bound_ft_glyph(fz_context *ctx, fz_font *font, int gid)
 	FT_BBox cbox;
 	FT_Matrix m;
 	FT_Vector v;
-	fz_rect *bounds = &font->bbox_table[gid];
+	fz_rect *bounds = get_gid_bbox(ctx, font, gid);
 
 	// TODO: refactor loading into fz_load_ft_glyph
 	// TODO: cache results
@@ -1425,15 +1443,16 @@ fz_bound_t3_glyph(fz_context *ctx, fz_font *font, int gid)
 {
 	fz_display_list *list;
 	fz_device *dev;
+	fz_rect *r = get_gid_bbox(ctx, font, gid);
 
 	list = font->t3lists[gid];
 	if (!list)
 	{
-		font->bbox_table[gid] = fz_empty_rect;
+		*r = fz_empty_rect;
 		return;
 	}
 
-	dev = fz_new_bbox_device(ctx, &font->bbox_table[gid]);
+	dev = fz_new_bbox_device(ctx, r);
 	fz_try(ctx)
 	{
 		fz_run_display_list(ctx, list, dev, font->t3matrix, fz_infinite_rect, NULL);
@@ -1450,20 +1469,15 @@ fz_bound_t3_glyph(fz_context *ctx, fz_font *font, int gid)
 
 	/* Update font bbox with glyph's computed bbox if the font bbox is invalid */
 	if (font->flags.invalid_bbox)
-		font->bbox = fz_union_rect(font->bbox, font->bbox_table[gid]);
+		font->bbox = fz_union_rect(font->bbox, *r);
 }
 
 void
 fz_prepare_t3_glyph(fz_context *ctx, fz_font *font, int gid)
 {
-	fz_buffer *contents;
 	fz_device *dev;
 	fz_rect d1_rect;
 
-	contents = font->t3procs[gid];
-	if (!contents)
-		return;
-
 	/* We've not already loaded this one! */
 	assert(font->t3lists[gid] == NULL);
 
@@ -1479,14 +1493,9 @@ fz_prepare_t3_glyph(fz_context *ctx, fz_font *font, int gid)
 			FZ_DEVFLAG_MITERLIMIT_UNDEFINED |
 			FZ_DEVFLAG_LINEWIDTH_UNDEFINED;
 
-	/* Avoid cycles in glyph content streams referring to the glyph itself.
-	 * Remember to restore the content stream below, regardless of exceptions
-	 * or a successful run of the glyph. */
-	font->t3procs[gid] = NULL;
-
 	fz_try(ctx)
 	{
-		font->t3run(ctx, font->t3doc, font->t3resources, contents, dev, fz_identity, NULL, NULL);
+		font->t3run(ctx, font->t3doc, font->t3resources, font->t3procs[gid], dev, fz_identity, NULL, NULL);
 		fz_close_device(ctx, dev);
 		font->t3flags[gid] = dev->flags;
 		d1_rect = dev->d1_rect;
@@ -1494,24 +1503,23 @@ fz_prepare_t3_glyph(fz_context *ctx, fz_font *font, int gid)
 	fz_always(ctx)
 	{
 		fz_drop_device(ctx, dev);
-		font->t3procs[gid] = contents;
 	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 	if (fz_display_list_is_empty(ctx, font->t3lists[gid]))
 	{
+		fz_rect *r = get_gid_bbox(ctx, font, gid);
 		/* If empty, no need for a huge bbox, especially as the logic
 		 * in the 'else if' can make it huge. */
-		font->bbox_table[gid].x0 = font->bbox.x0;
-		font->bbox_table[gid].y0 = font->bbox.y0;
-		font->bbox_table[gid].x1 = font->bbox.x0 + .00001f;
-		font->bbox_table[gid].y1 = font->bbox.y0 + .00001f;
+		r->x0 = font->flags.invalid_bbox ? 0 : font->bbox.x0;
+		r->y0 = font->flags.invalid_bbox ? 0 : font->bbox.y0;
+		r->x1 = r->x0 + .00001f;
+		r->y1 = r->y0 + .00001f;
 	}
 	else if (font->t3flags[gid] & FZ_DEVFLAG_BBOX_DEFINED)
 	{
-		assert(font->bbox_table != NULL);
-		assert(font->glyph_count > gid);
-		font->bbox_table[gid] = fz_transform_rect(d1_rect, font->t3matrix);
+		fz_rect *r = get_gid_bbox(ctx, font, gid);
+		*r = fz_transform_rect(d1_rect, font->t3matrix);
 
 		if (font->flags.invalid_bbox || !fz_contains_rect(font->bbox, d1_rect))
 		{
@@ -1633,15 +1641,10 @@ void
 fz_render_t3_glyph_direct(fz_context *ctx, fz_device *dev, fz_font *font, int gid, fz_matrix trm, void *gstate, fz_default_colorspaces *def_cs)
 {
 	fz_matrix ctm;
-	void *contents;
 
 	if (gid < 0 || gid > 255)
 		return;
 
-	contents = font->t3procs[gid];
-	if (!contents)
-		return;
-
 	if (font->t3flags[gid] & FZ_DEVFLAG_MASK)
 	{
 		if (font->t3flags[gid] & FZ_DEVFLAG_COLOR)
@@ -1652,30 +1655,19 @@ fz_render_t3_glyph_direct(fz_context *ctx, fz_device *dev, fz_font *font, int gi
 		fz_warn(ctx, "type3 glyph doesn't specify masked or colored");
 	}
 
-	/* Avoid cycles in glyph content streams referring to the glyph itself.
-	 * Remember to restore the content stream below, regardless of exceptions
-	 * or a successful run of the glyph. */
-	font->t3procs[gid] = NULL;
-
-	fz_try(ctx)
-	{
-		ctm = fz_concat(font->t3matrix, trm);
-		font->t3run(ctx, font->t3doc, font->t3resources, contents, dev, ctm, gstate, def_cs);
-	}
-	fz_always(ctx)
-		font->t3procs[gid] = contents;
-	fz_catch(ctx)
-		fz_rethrow(ctx);
+	ctm = fz_concat(font->t3matrix, trm);
+	font->t3run(ctx, font->t3doc, font->t3resources, font->t3procs[gid], dev, ctm, gstate, def_cs);
 }
 
 fz_rect
 fz_bound_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix trm)
 {
 	fz_rect rect;
-	if (font->bbox_table && gid < font->glyph_count && gid >= 0)
+	fz_rect *r = get_gid_bbox(ctx, font, gid);
+	if (r)
 	{
 		/* If the bbox is infinite or empty, distrust it */
-		if (fz_is_infinite_rect(font->bbox_table[gid]) || fz_is_empty_rect(font->bbox_table[gid]))
+		if (fz_is_infinite_rect(*r) || fz_is_empty_rect(*r))
 		{
 			/* Get the real size from the glyph */
 			if (font->ft_face)
@@ -1685,19 +1677,19 @@ fz_bound_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix trm)
 			else
 				/* If we can't get a real size, fall back to the font
 				 * bbox. */
-				font->bbox_table[gid] = font->bbox;
+				*r = font->bbox;
 			/* If the real size came back as empty, then store it as
 			 * a very small rectangle to avoid us calling this same
 			 * check every time. */
-			if (fz_is_empty_rect(font->bbox_table[gid]))
+			if (fz_is_empty_rect(*r))
 			{
-				font->bbox_table[gid].x0 = 0;
-				font->bbox_table[gid].y0 = 0;
-				font->bbox_table[gid].x1 = 0.0000001f;
-				font->bbox_table[gid].y1 = 0.0000001f;
+				r->x0 = 0;
+				r->y0 = 0;
+				r->x1 = 0.0000001f;
+				r->y1 = 0.0000001f;
 			}
 		}
-		rect = font->bbox_table[gid];
+		rect = *r;
 	}
 	else
 	{
@@ -1803,24 +1795,41 @@ fz_advance_glyph(fz_context *ctx, fz_font *font, int gid, int wmode)
 	{
 		if (wmode)
 			return fz_advance_ft_glyph(ctx, font, gid, 1);
-		if (gid >= 0 && gid < font->glyph_count && gid < MAX_ADVANCE_CACHE)
+		if (gid >= 0 && gid < font->glyph_count)
 		{
 			float f;
+			int block = gid>>8;
 			fz_lock(ctx, FZ_LOCK_FREETYPE);
 			if (!font->advance_cache)
 			{
-				int i;
+				int n = (font->glyph_count+255)/256;
 				fz_try(ctx)
-					font->advance_cache = Memento_label(fz_malloc_array(ctx, font->glyph_count, float), "font_advance_cache");
+					font->advance_cache = Memento_label(fz_malloc_array(ctx, n, float *), "font_advance_cache");
 				fz_catch(ctx)
 				{
 					fz_unlock(ctx, FZ_LOCK_FREETYPE);
 					fz_rethrow(ctx);
 				}
-				for (i = 0; i < font->glyph_count; ++i)
-					font->advance_cache[i] = fz_advance_ft_glyph_aux(ctx, font, i, 0, 1);
+				memset(font->advance_cache, 0, n * sizeof(float *));
 			}
-			f = font->advance_cache[gid];
+			if (!font->advance_cache[block])
+			{
+				int i, n;
+				fz_try(ctx)
+					font->advance_cache[block] = Memento_label(fz_malloc_array(ctx, 256, float), "font_advance_cache");
+				fz_catch(ctx)
+				{
+					fz_unlock(ctx, FZ_LOCK_FREETYPE);
+					fz_rethrow(ctx);
+				}
+				n = (block<<8)+256;
+				if (n > font->glyph_count)
+					n = font->glyph_count;
+				n -= (block<<8);
+				for (i = 0; i < n; ++i)
+					font->advance_cache[block][i] = fz_advance_ft_glyph_aux(ctx, font, (block<<8)+i, 0, 1);
+			}
+			f = font->advance_cache[block][gid & 255];
 			fz_unlock(ctx, FZ_LOCK_FREETYPE);
 			return f;
 		}
