@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2022 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -63,6 +63,7 @@ fz_save_pixmap_as_png(fz_context *ctx, fz_pixmap *pixmap, const char *filename)
 		writer = fz_new_png_band_writer(ctx, out);
 		fz_write_header(ctx, writer, pixmap->w, pixmap->h, pixmap->n, pixmap->alpha, pixmap->xres, pixmap->yres, 0, pixmap->colorspace, pixmap->seps);
 		fz_write_band(ctx, writer, pixmap->stride, pixmap->h, pixmap->samples);
+		fz_close_band_writer(ctx, writer);
 		fz_close_output(ctx, out);
 	}
 	fz_always(ctx)
@@ -90,6 +91,7 @@ fz_write_pixmap_as_png(fz_context *ctx, fz_output *out, const fz_pixmap *pixmap)
 	{
 		fz_write_header(ctx, writer, pixmap->w, pixmap->h, pixmap->n, pixmap->alpha, pixmap->xres, pixmap->yres, 0, pixmap->colorspace, pixmap->seps);
 		fz_write_band(ctx, writer, pixmap->stride, pixmap->h, pixmap->samples);
+		fz_close_band_writer(ctx, writer);
 	}
 	fz_always(ctx)
 	{
@@ -169,6 +171,8 @@ png_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *cs)
 
 	if (writer->super.s != 0)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "PNGs cannot contain spot colors");
+	if (fz_colorspace_type(ctx, cs) == FZ_COLORSPACE_BGR)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "pixmap can not be bgr");
 
 	/* Treat alpha only as greyscale */
 	if (n == 1 && alpha)
