@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #define _LARGEFILE_SOURCE
 #ifndef _FILE_OFFSET_BITS
 #define _FILE_OFFSET_BITS 64
@@ -11,6 +33,7 @@
 #include <string.h>
 #ifdef _WIN32
 #include <io.h>
+#include <windows.h>
 #else
 #include <unistd.h>
 #endif
@@ -75,6 +98,48 @@ fz_output *fz_stderr(fz_context *ctx)
 	return &fz_stderr_global;
 }
 
+#ifdef _WIN32
+static void
+stdods_write(fz_context *ctx, void *opaque, const void *buffer, size_t count)
+{
+	unsigned char *buf = fz_malloc(ctx, count+1);
+
+	memcpy(buf, buffer, count);
+	buf[count] = 0;
+	OutputDebugStringA(buf);
+	fz_free(ctx, buf);
+}
+
+static fz_output fz_stdods_global = {
+	NULL,
+	stdods_write,
+	NULL,
+	NULL,
+	NULL,
+};
+
+fz_output *fz_stdods(fz_context *ctx)
+{
+	return &fz_stdods_global;
+}
+#endif
+
+fz_output *fz_stddbg(fz_context *ctx)
+{
+	if (ctx->stddbg)
+		return ctx->stddbg;
+
+	return fz_stderr(ctx);
+}
+
+void fz_set_stddbg(fz_context *ctx, fz_output *out)
+{
+	if (ctx == NULL)
+		return;
+
+	ctx->stddbg = out;
+}
+
 static void
 file_seek(fz_context *ctx, void *opaque, int64_t off, int whence)
 {
@@ -125,9 +190,17 @@ static void file_truncate(fz_context *ctx, void *opaque)
 	fflush(file);
 
 #ifdef _WIN32
-	_chsize_s(fileno(file), ftell(file));
+	{
+		__int64 pos = _ftelli64(file);
+		if (pos >= 0)
+			_chsize_s(fileno(file), pos);
+	}
 #else
-	ftruncate(fileno(file), ftell(file));
+	{
+		off_t pos = ftello(file);
+		if (pos >= 0)
+			(void)ftruncate(fileno(file), pos);
+	}
 #endif
 }
 
@@ -172,9 +245,19 @@ fz_new_output_with_path(fz_context *ctx, const char *filename, int append)
 	FILE *file;
 	fz_output *out;
 
+	if (filename == NULL)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "no output to write to");
+
 	if (!strcmp(filename, "/dev/null") || !fz_strcasecmp(filename, "nul:"))
 		return fz_new_output(ctx, 0, NULL, null_write, NULL, NULL);
 
+	/* If <append> is false, we use fopen()'s 'x' flag to force an error if
+	 * some other process creates the file immediately after we have removed
+	 * it - this avoids vunerability where a less-privilege process can create
+	 * a link and get us to overwrite a different file. See:
+	 * 	https://bugs.ghostscript.com/show_bug.cgi?id=701797
+	 * 	http://www.open-std.org/jtc1/sc22//WG14/www/docs/n1339.pdf
+	 */
 #ifdef _WIN32
 	/* Ensure we create a brand new file. We don't want to clobber our old file. */
 	if (!append)
@@ -183,7 +266,11 @@ fz_new_output_with_path(fz_context *ctx, const char *filename, int append)
 			if (errno != ENOENT)
 				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot remove file '%s': %s", filename, strerror(errno));
 	}
-	file = fz_fopen_utf8(filename, append ? "rb+" : "wb+");
+#if defined(__MINGW32__) || defined(__MINGW64__)
+	file = fz_fopen_utf8(filename, append ? "rb+" : "wb+"); /* 'x' flag not suported. */
+#else
+	file = fz_fopen_utf8(filename, append ? "rb+" : "wb+x");
+#endif
 	if (append)
 	{
 		if (file == NULL)
@@ -199,7 +286,7 @@ fz_new_output_with_path(fz_context *ctx, const char *filename, int append)
 			if (errno != ENOENT)
 				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot remove file '%s': %s", filename, strerror(errno));
 	}
-	file = fopen(filename, append ? "rb+" : "wb+");
+	file = fopen(filename, append ? "rb+" : "wb+x");
 	if (file == NULL && append)
 		file = fopen(filename, "wb+");
 #endif
@@ -611,6 +698,15 @@ void fz_write_band(fz_context *ctx, fz_band_writer *writer, int stride, int band
 	}
 }
 
+void fz_close_band_writer(fz_context *ctx, fz_band_writer *writer)
+{
+	if (writer == NULL)
+		return;
+	if (writer->close != NULL)
+		writer->close(ctx, writer);
+	writer->close = NULL;
+}
+
 void fz_drop_band_writer(fz_context *ctx, fz_band_writer *writer)
 {
 	if (writer == NULL)
