@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2022 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -36,6 +36,8 @@ struct pdf_js
 	pdf_document *doc;
 	pdf_obj *form;
 	js_State *imp;
+	pdf_js_console *console;
+	void *console_user;
 };
 
 FZ_NORETURN static void rethrow(pdf_js *js)
@@ -72,20 +74,78 @@ static pdf_js *unpack_arguments(js_State *J, ...)
 
 static void app_alert(js_State *J)
 {
-	pdf_js *js = unpack_arguments(J, "cMsg", "nIcon", "nType", "cTitle", 0);
+	pdf_js *js = unpack_arguments(J, "cMsg", "nIcon", "nType", "cTitle", "oDoc", "oCheckbox", NULL);
 	pdf_alert_event evt;
 
+	/* TODO: Currently we do not support app.openDoc() in javascript actions, hence
+	oDoc can only point to the current document (or not be passed). When mupdf
+	supports opening other documents oDoc must be converted to a pdf_document * that
+	can be passed to the callback. In the mean time, we just pas the current document.
+	*/
+	evt.doc = js->doc;
+
 	evt.message = js_tostring(J, 1);
 	evt.icon_type = js_tointeger(J, 2);
 	evt.button_group_type = js_tointeger(J, 3);
-	evt.title = js_isdefined(J, 4) ? js_tostring(J, 4) : "PDF Alert";
-	evt.button_pressed = 0; /* WIP WIP WIP IS THIS CORRECT? */
+	evt.title = js_isdefined(J, 4) ? js_tostring(J, 4) : "PDF alert";
+
+	evt.has_check_box = 0;
+	evt.check_box_message = NULL;
+	evt.initially_checked = 0;
+	evt.finally_checked = 0;
+
+	if (js_isobject(J, 6))
+	{
+		evt.has_check_box = 1;
+		evt.check_box_message = "Do not show this message again";
+		if (js_hasproperty(J, 6, "cMsg"))
+		{
+			if (js_iscoercible(J, -1))
+				evt.check_box_message = js_tostring(J, -1);
+			js_pop(J, 1);
+		}
+		if (js_hasproperty(J, 6, "bInitialValue"))
+		{
+			evt.initially_checked = js_tointeger(J, -1);
+			js_pop(J, 1);
+		}
+		if (js_hasproperty(J, 6, "bAfterValue"))
+		{
+			evt.finally_checked = js_tointeger(J, -1);
+			js_pop(J, 1);
+		}
+	}
+
+	/* These are the default buttons automagically "pressed"
+	when the dialog box window is closed in Acrobat. */
+	switch (evt.button_group_type)
+	{
+	default:
+	case PDF_ALERT_BUTTON_GROUP_OK:
+		evt.button_pressed = PDF_ALERT_BUTTON_OK;
+		break;
+	case PDF_ALERT_BUTTON_GROUP_OK_CANCEL:
+		evt.button_pressed = PDF_ALERT_BUTTON_CANCEL;
+		break;
+	case PDF_ALERT_BUTTON_GROUP_YES_NO:
+		evt.button_pressed = PDF_ALERT_BUTTON_YES;
+		break;
+	case PDF_ALERT_BUTTON_GROUP_YES_NO_CANCEL:
+		evt.button_pressed = PDF_ALERT_BUTTON_CANCEL;
+		break;
+	}
 
 	fz_try(js->ctx)
 		pdf_event_issue_alert(js->ctx, js->doc, &evt);
 	fz_catch(js->ctx)
 		rethrow(js);
 
+	if (js_isobject(J, 6))
+	{
+		js_pushboolean(js->imp, evt.finally_checked);
+		js_setproperty(js->imp, 6, "bAfterValue");
+	}
+
 	js_pushnumber(J, evt.button_pressed);
 }
 
@@ -307,6 +367,26 @@ static void field_setValue(js_State *J)
 		rethrow(js);
 }
 
+static void field_getType(js_State *J)
+{
+	pdf_js *js = js_getcontext(J);
+	pdf_obj *field = js_touserdata(J, 0, "Field");
+	const char *type;
+
+	fz_try(js->ctx)
+		type = pdf_field_type_string(js->ctx, field);
+	fz_catch(js->ctx)
+		rethrow(js);
+
+	js_pushstring(J, type);
+}
+
+static void field_setType(js_State *J)
+{
+	pdf_js *js = js_getcontext(J);
+	fz_warn(js->ctx, "Unexpected call to field_setType");
+}
+
 static void doc_getField(js_State *J)
 {
 	pdf_js *js = js_getcontext(J);
@@ -343,6 +423,93 @@ static void doc_setNumPages(js_State *J)
 	fz_warn(js->ctx, "Unexpected call to doc_setNumPages");
 }
 
+static void doc_getMetaString(js_State *J, const char *key)
+{
+	pdf_js *js = js_getcontext(J);
+	char buf[256];
+	int ret;
+
+	fz_try(js->ctx)
+		ret = fz_lookup_metadata(js->ctx, &js->doc->super, key, buf, nelem(buf)) > 0;
+	fz_catch(js->ctx)
+		rethrow(js);
+
+	if (ret > 0)
+		js_pushstring(J, buf);
+	else
+		js_pushundefined(J);
+}
+
+static void doc_setMetaString(js_State *J, const char *key)
+{
+	pdf_js *js = js_getcontext(J);
+	const char *value = js_tostring(J, 1);
+	fz_set_metadata(js->ctx, &js->doc->super, key, value);
+}
+
+static void doc_getMetaDate(js_State *J, const char *key)
+{
+	pdf_js *js = js_getcontext(J);
+	char buf[256];
+	int ret;
+	double time;
+
+	fz_try(js->ctx)
+	{
+		ret = fz_lookup_metadata(js->ctx, &js->doc->super, key, buf, nelem(buf)) > 0;
+		if (ret > 0)
+			time = pdf_parse_date(js->ctx, buf);
+	}
+	fz_catch(js->ctx)
+		rethrow(js);
+
+	if (ret > 0)
+	{
+		js_getglobal(J, "Date");
+		js_pushnumber(J, time * 1000);
+		js_construct(J, 1);
+	}
+	else
+		js_pushundefined(J);
+}
+
+static void doc_setMetaDate(js_State *J, const char *key)
+{
+	pdf_js *js = js_getcontext(J);
+	int64_t time;
+	char value[40];
+
+	// Coerce the argument into a date object and extract the time value.
+	js_getglobal(J, "Date");
+	js_copy(J, 1);
+	js_construct(J, 1);
+	time = js_tonumber(J, -1) / 1000;
+	js_pop(J, 1);
+
+	fz_try(js->ctx)
+		if (pdf_format_date(js->ctx, time, value, nelem(value)))
+			fz_set_metadata(js->ctx, &js->doc->super, key, value);
+	fz_catch(js->ctx)
+		rethrow(js);
+}
+
+static void doc_getAuthor(js_State *J) { doc_getMetaString(J, FZ_META_INFO_AUTHOR); }
+static void doc_setAuthor(js_State *J) { doc_setMetaString(J, FZ_META_INFO_AUTHOR); }
+static void doc_getTitle(js_State *J) { doc_getMetaString(J, FZ_META_INFO_TITLE); }
+static void doc_setTitle(js_State *J) { doc_setMetaString(J, FZ_META_INFO_TITLE); }
+static void doc_getSubject(js_State *J) { doc_getMetaString(J, FZ_META_INFO_SUBJECT); }
+static void doc_setSubject(js_State *J) { doc_setMetaString(J, FZ_META_INFO_SUBJECT); }
+static void doc_getKeywords(js_State *J) { doc_getMetaString(J, FZ_META_INFO_KEYWORDS); }
+static void doc_setKeywords(js_State *J) { doc_setMetaString(J, FZ_META_INFO_KEYWORDS); }
+static void doc_getCreator(js_State *J) { doc_getMetaString(J, FZ_META_INFO_CREATOR); }
+static void doc_setCreator(js_State *J) { doc_setMetaString(J, FZ_META_INFO_CREATOR); }
+static void doc_getProducer(js_State *J) { doc_getMetaString(J, FZ_META_INFO_PRODUCER); }
+static void doc_setProducer(js_State *J) { doc_setMetaString(J, FZ_META_INFO_PRODUCER); }
+static void doc_getCreationDate(js_State *J) { doc_getMetaDate(J, FZ_META_INFO_CREATIONDATE); }
+static void doc_setCreationDate(js_State *J) { doc_setMetaDate(J, FZ_META_INFO_CREATIONDATE); }
+static void doc_getModDate(js_State *J) { doc_getMetaDate(J, FZ_META_INFO_MODIFICATIONDATE); }
+static void doc_setModDate(js_State *J) { doc_setMetaDate(J, FZ_META_INFO_MODIFICATIONDATE); }
+
 static void doc_resetForm(js_State *J)
 {
 	pdf_js *js = js_getcontext(J);
@@ -389,7 +556,7 @@ static void doc_print(js_State *J)
 
 static void doc_mailDoc(js_State *J)
 {
-	pdf_js *js = unpack_arguments(J, "bUI", "cTo", "cCc", "cBcc", "cSubject", "cMessage", 0);
+	pdf_js *js = unpack_arguments(J, "bUI", "cTo", "cCc", "cBcc", "cSubject", "cMessage", NULL);
 	pdf_mail_doc_event evt;
 
 	evt.ask_user = js_isdefined(J, 1) ? js_toboolean(J, 1) : 1;
@@ -417,13 +584,41 @@ static void doc_calculateNow(js_State *J)
 static void console_println(js_State *J)
 {
 	pdf_js *js = js_getcontext(J);
-	int i, top = js_gettop(J);
-	for (i = 1; i < top; ++i) {
-		const char *s = js_tostring(J, i);
-		if (i > 1) fz_write_byte(js->ctx, fz_stddbg(js->ctx), ' ');
-		fz_write_string(js->ctx, fz_stddbg(js->ctx), s);
+	if (js->console && js->console->write)
+	{
+		int i, top = js_gettop(J);
+		js->console->write(js->console_user, "\n");
+		for (i = 1; i < top; ++i) {
+			const char *s = js_tostring(J, i);
+			if (i > 1)
+				js->console->write(js->console_user, " ");
+			js->console->write(js->console_user, s);
+		}
 	}
-	fz_write_byte(js->ctx, fz_stddbg(js->ctx), '\n');
+	js_pushboolean(J, 1);
+}
+
+static void console_clear(js_State *J)
+{
+	pdf_js *js = js_getcontext(J);
+	if (js->console && js->console->clear)
+		js->console->clear(js->console_user);
+	js_pushundefined(J);
+}
+
+static void console_show(js_State *J)
+{
+	pdf_js *js = js_getcontext(J);
+	if (js->console && js->console->show)
+		js->console->show(js->console_user);
+	js_pushundefined(J);
+}
+
+static void console_hide(js_State *J)
+{
+	pdf_js *js = js_getcontext(J);
+	if (js->console && js->console->hide)
+		js->console->hide(js->console_user);
 	js_pushundefined(J);
 }
 
@@ -712,7 +907,7 @@ static void declare_dom(pdf_js *js)
 #endif
 		js_defproperty(J, -2, "app.platform", JS_READONLY | JS_DONTENUM | JS_DONTCONF);
 
-		addmethod(J, "app.alert", app_alert, 4);
+		addmethod(J, "app.alert", app_alert, 6);
 		addmethod(J, "app.execMenuItem", app_execMenuItem, 1);
 		addmethod(J, "app.launchURL", app_launchURL, 2);
 	}
@@ -722,6 +917,7 @@ static void declare_dom(pdf_js *js)
 	js_newobject(J);
 	{
 		addproperty(J, "Field.value", field_getValue, field_setValue);
+		addproperty(J, "Field.type", field_getType, field_setType);
 		addproperty(J, "Field.borderStyle", field_getBorderStyle, field_setBorderStyle);
 		addproperty(J, "Field.textColor", field_getTextColor, field_setTextColor);
 		addproperty(J, "Field.fillColor", field_getFillColor, field_setFillColor);
@@ -735,6 +931,9 @@ static void declare_dom(pdf_js *js)
 	js_newobject(J);
 	{
 		addmethod(J, "console.println", console_println, 1);
+		addmethod(J, "console.clear", console_clear, 0);
+		addmethod(J, "console.show", console_show, 0);
+		addmethod(J, "console.hide", console_hide, 0);
 	}
 	js_defglobal(J, "console", JS_READONLY | JS_DONTCONF | JS_DONTENUM);
 
@@ -743,6 +942,14 @@ static void declare_dom(pdf_js *js)
 	js_pushglobal(J);
 	{
 		addproperty(J, "Doc.numPages", doc_getNumPages, doc_setNumPages);
+		addproperty(J, "Doc.author", doc_getAuthor, doc_setAuthor);
+		addproperty(J, "Doc.title", doc_getTitle, doc_setTitle);
+		addproperty(J, "Doc.subject", doc_getSubject, doc_setSubject);
+		addproperty(J, "Doc.keywords", doc_getKeywords, doc_setKeywords);
+		addproperty(J, "Doc.creator", doc_getCreator, doc_setCreator);
+		addproperty(J, "Doc.producer", doc_getProducer, doc_setProducer);
+		addproperty(J, "Doc.creationDate", doc_getCreationDate, doc_setCreationDate);
+		addproperty(J, "Doc.modDate", doc_getModDate, doc_setModDate);
 		addmethod(J, "Doc.getField", doc_getField, 1);
 		addmethod(J, "Doc.resetForm", doc_resetForm, 0);
 		addmethod(J, "Doc.calculateNow", doc_calculateNow, 0);
@@ -779,6 +986,8 @@ void pdf_drop_js(fz_context *ctx, pdf_js *js)
 {
 	if (js)
 	{
+		if (js->console && js->console->drop)
+			js->console->drop(js->console, js->console_user);
 		js_freestate(js->imp);
 		fz_free(ctx, js);
 	}
@@ -789,6 +998,26 @@ static void *pdf_js_alloc(void *actx, void *ptr, int n)
 	return fz_realloc_no_throw(actx, ptr, n);
 }
 
+static void default_js_console_clear(void *user)
+{
+	fz_context *ctx = user;
+	fz_write_string(ctx, fz_stddbg(ctx), "--- clear console ---\n");
+}
+
+static void default_js_console_write(void *user, const char *message)
+{
+	fz_context *ctx = user;
+	fz_write_string(ctx, fz_stddbg(ctx), message);
+}
+
+static pdf_js_console default_js_console = {
+	NULL,
+	NULL,
+	NULL,
+	default_js_console_clear,
+	default_js_console_write,
+};
+
 static pdf_js *pdf_new_js(fz_context *ctx, pdf_document *doc)
 {
 	pdf_js *js = fz_malloc_struct(ctx, pdf_js);
@@ -813,6 +1042,9 @@ static pdf_js *pdf_new_js(fz_context *ctx, pdf_document *doc)
 		/* Also set our pdf_js context, so we can retrieve it in callbacks. */
 		js_setcontext(js->imp, js);
 
+		js->console = &default_js_console;
+		js->console_user = js->ctx;
+
 		declare_dom(js);
 		preload_helpers(js);
 	}
@@ -852,7 +1084,7 @@ static void pdf_js_load_document_level(pdf_js *js)
 				fz_snprintf(buf, sizeof buf, "%d", pdf_to_num(ctx, code));
 			else
 				fz_snprintf(buf, sizeof buf, "Root/Names/JavaScript/Names/%d/JS", (i+1)*2);
-			pdf_js_execute(js, buf, codebuf);
+			pdf_js_execute(js, buf, codebuf, NULL);
 			fz_free(ctx, codebuf);
 		}
 	}
@@ -985,38 +1217,62 @@ char *pdf_js_event_value(pdf_js *js)
 	return value;
 }
 
-void pdf_js_execute(pdf_js *js, const char *name, const char *source)
+void pdf_js_execute(pdf_js *js, const char *name, const char *source, char **result)
 {
 	fz_context *ctx;
+	js_State *J;
 
 	if (!js)
 		return;
 
 	ctx = js->ctx;
-	pdf_begin_implicit_operation(js->ctx, js->doc);
+	J = js->imp;
+
+	pdf_begin_implicit_operation(ctx, js->doc);
 	fz_try(ctx)
 	{
-		if (js_ploadstring(js->imp, name, source))
-		{
-			fz_warn(ctx, "%s", js_trystring(js->imp, -1, "Error"));
-			break;
-		}
-		js_pushundefined(js->imp);
-		if (js_pcall(js->imp, 0))
-		{
-			fz_warn(ctx, "%s", js_trystring(js->imp, -1, "Error"));
-			break;
+		if (js_ploadstring(J, name, source)) {
+			if (result)
+				*result = fz_strdup(ctx, js_trystring(J, -1, "Error"));
+			js_pop(J, 1);
+		} else {
+			js_pushundefined(J);
+			if (js_pcall(J, 0)) {
+				if (result)
+					*result = fz_strdup(ctx, js_trystring(J, -1, "Error"));
+				js_pop(J, 1);
+			} else {
+				if (result)
+					*result = fz_strdup(ctx, js_tryrepr(J, -1, "can't convert to string"));
+				js_pop(J, 1);
+			}
 		}
 	}
 	fz_always(ctx)
 	{
-		js_pop(js->imp, 1);
-		pdf_end_operation(js->ctx, js->doc);
+		pdf_end_operation(ctx, js->doc);
 	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 }
 
+pdf_js_console *pdf_js_get_console(fz_context *ctx, pdf_document *doc)
+{
+	return doc->js ? doc->js->console : NULL;
+}
+
+void pdf_js_set_console(fz_context *ctx, pdf_document *doc, pdf_js_console *console, void *user)
+{
+	if (doc->js)
+	{
+		if (doc->js->console && doc->js->console->drop)
+			doc->js->console->drop(doc->js->console, doc->js->console_user);
+
+		doc->js->console = console;
+		doc->js->console_user = user;
+	}
+}
+
 void pdf_enable_js(fz_context *ctx, pdf_document *doc)
 {
 	if (!doc->js)
@@ -1048,7 +1304,9 @@ void pdf_js_event_init_keystroke(pdf_js *js, pdf_obj *target, pdf_keystroke_even
 int pdf_js_event_result_keystroke(pdf_js *js, pdf_keystroke_event *evt) { return 1; }
 int pdf_js_event_result(pdf_js *js) { return 1; }
 char *pdf_js_event_value(pdf_js *js) { return ""; }
-void pdf_js_execute(pdf_js *js, const char *name, const char *source) { }
+void pdf_js_execute(pdf_js *js, const char *name, const char *source, char **result) { }
 int pdf_js_event_result_validate(pdf_js *js, char **newvalue) { *newvalue=NULL; return 1; }
+pdf_js_console *pdf_js_get_console(fz_context *ctx, pdf_document *doc) { return NULL; }
+void pdf_js_set_console(fz_context *ctx, pdf_document *doc, pdf_js_console *console, void *user) { }
 
 #endif /* FZ_ENABLE_JS */
