@@ -26,7 +26,6 @@
 #include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
-#include <time.h>
 
 #define PDF_MAKE_NAME(STRING,NAME) STRING,
 static const char *PDF_NAME_LIST[] = {
@@ -1336,6 +1335,10 @@ void pdf_deserialise_journal(fz_context *ctx, pdf_document *doc, fz_stream *stm)
 	}
 
 	doc->file_size = file_size;
+	/* We're about to make the last xref an incremental one. All incremental
+	 * ones MUST be solid, but the snapshot might not have saved it as such,
+	 * so solidify it now. */
+	pdf_ensure_solid_xref(ctx, doc, pdf_xref_len(ctx, doc));
 	doc->num_incremental_sections = nis;
 
 	if (nis > 0)
@@ -1396,7 +1399,7 @@ static void prepare_object_for_alteration(fz_context *ctx, pdf_obj *obj, pdf_obj
 		parent_num == 0 while an object is being parsed from the file.
 		No further action is necessary.
 	*/
-	if (parent == 0 || doc->save_in_progress || doc->repair_attempted)
+	if (parent == 0 || doc->save_in_progress || doc->repair_in_progress)
 		return;
 
 	if (doc->journal && doc->journal->nesting == 0)
@@ -1638,19 +1641,12 @@ pdf_obj *pdf_new_matrix(fz_context *ctx, pdf_document *doc, fz_matrix mtx)
 	return arr;
 }
 
+
 pdf_obj *pdf_new_date(fz_context *ctx, pdf_document *doc, int64_t time)
 {
 	char s[40];
-	time_t secs = time;
-#ifdef _POSIX_SOURCE
-	struct tm tmbuf, *tm = gmtime_r(&secs, &tmbuf);
-#else
-	struct tm *tm = gmtime(&secs);
-#endif
-
-	if (time < 0 || !tm || !strftime(s, nelem(s), "D:%Y%m%d%H%M%SZ", tm))
+	if (!pdf_format_date(ctx, time, s, nelem(s)))
 		return NULL;
-
 	return pdf_new_string(ctx, s, strlen(s));
 }
 
@@ -2412,6 +2408,123 @@ pdf_unmark_obj(fz_context *ctx, pdf_obj *obj)
 	obj->flags &= ~PDF_FLAGS_MARKED;
 }
 
+int
+pdf_cycle(fz_context *ctx, pdf_cycle_list *here, pdf_cycle_list *up, pdf_obj *obj)
+{
+	int num = pdf_to_num(ctx, obj);
+	if (num > 0)
+	{
+		pdf_cycle_list *x = up;
+		while (x)
+		{
+			if (x->num == num)
+				return 1;
+			x = x->up;
+		}
+	}
+	here->up = up;
+	here->num = num;
+	return 0;
+}
+
+pdf_mark_bits *
+pdf_new_mark_bits(fz_context *ctx, pdf_document *doc)
+{
+	int n = pdf_xref_len(ctx, doc);
+	int nb = (n + 7) >> 3;
+	pdf_mark_bits *marks = fz_calloc(ctx, offsetof(pdf_mark_bits, bits) + nb, 1);
+	marks->len = n;
+	return marks;
+}
+
+void
+pdf_drop_mark_bits(fz_context *ctx, pdf_mark_bits *marks)
+{
+	fz_free(ctx, marks);
+}
+
+void pdf_mark_bits_reset(fz_context *ctx, pdf_mark_bits *marks)
+{
+	memset(marks->bits, 0, (marks->len + 7) >> 3);
+}
+
+int pdf_mark_bits_set(fz_context *ctx, pdf_mark_bits *marks, pdf_obj *obj)
+{
+	int num = pdf_to_num(ctx, obj);
+	if (num > 0 && num < marks->len)
+	{
+		int x = num >> 3;
+		int m = 1 << (num & 7);
+		if (marks->bits[x] & m)
+			return 1;
+		marks->bits[x] |= m;
+	}
+	return 0;
+}
+
+void pdf_mark_bits_clear(fz_context *ctx, pdf_mark_bits *marks, pdf_obj *obj)
+{
+	int num = pdf_to_num(ctx, obj);
+	if (num > 0 && num < marks->len)
+	{
+		int x = num >> 3;
+		int m = 0xff ^ (1 << (num & 7));
+		marks->bits[x] &= m;
+	}
+}
+
+int
+pdf_mark_list_push(fz_context *ctx, pdf_mark_list *marks, pdf_obj *obj)
+{
+	int num = pdf_to_num(ctx, obj);
+	int i;
+
+	if (num > 0)
+	{
+		/* Note: this is slow, if the mark list is expected to be big use pdf_mark_bits instead! */
+		for (i = 0; i < marks->len; ++i)
+			if (marks->list[i] == num)
+				return 1;
+	}
+
+	if (marks->len == marks->max)
+	{
+		int newsize = marks->max << 1;
+		if (marks->list == marks->local_list)
+			marks->list = fz_malloc_array(ctx, newsize, int);
+		else
+			marks->list = fz_realloc_array(ctx, marks->list, newsize, int);
+		marks->max = newsize;
+	}
+
+	marks->list[marks->len++] = num;
+	return 0;
+}
+
+void
+pdf_mark_list_pop(fz_context *ctx, pdf_mark_list *marks)
+{
+	--marks->len;
+}
+
+void
+pdf_mark_list_init(fz_context *ctx, pdf_mark_list *marks)
+{
+	marks->len = 0;
+	marks->max = nelem(marks->local_list);
+	marks->list = marks->local_list;
+}
+
+void
+pdf_mark_list_free(fz_context *ctx, pdf_mark_list *marks)
+{
+	if (marks->list != marks->local_list)
+		fz_free(ctx, marks->list);
+	marks->len = 0;
+	marks->max = 0;
+	marks->list = NULL;
+}
+
 void
 pdf_set_obj_memo(fz_context *ctx, pdf_obj *obj, int bit, int memo)
 {
@@ -3033,100 +3146,50 @@ int pdf_obj_refs(fz_context *ctx, pdf_obj *obj)
 
 /* Convenience functions */
 
+static pdf_obj *
+pdf_dict_get_inheritable_imp(fz_context *ctx, pdf_obj *node, pdf_obj *key, int depth, pdf_cycle_list *cycle_up)
+{
+	pdf_cycle_list cycle;
+	pdf_obj *val = pdf_dict_get(ctx, node, key);
+	if (val)
+		return val;
+	if (pdf_cycle(ctx, &cycle, cycle_up, node))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in tree (parents)");
+	if (depth > 100)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "too much recursion in tree (parents)");
+	node = pdf_dict_get(ctx, node, PDF_NAME(Parent));
+	if (node)
+		return pdf_dict_get_inheritable_imp(ctx, node, key, depth + 1, &cycle);
+	return NULL;
+}
+
 pdf_obj *
 pdf_dict_get_inheritable(fz_context *ctx, pdf_obj *node, pdf_obj *key)
 {
-	pdf_obj *node2 = node;
-	pdf_obj *val = NULL;
-	pdf_obj *marked = NULL;
+	return pdf_dict_get_inheritable_imp(ctx, node, key, 0, NULL);
+}
 
-	fz_var(node);
-	fz_var(marked);
-	fz_try(ctx)
-	{
-		do
-		{
-			val = pdf_dict_get(ctx, node, key);
-			if (val)
-				break;
-			if (pdf_mark_obj(ctx, node))
-				fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in tree (parents)");
-			marked = node;
-			node = pdf_dict_get(ctx, node, PDF_NAME(Parent));
-		}
-		while (node);
-	}
-	fz_always(ctx)
-	{
-		/* We assume that if we have marked an object, without an exception
-		 * being thrown, that we can always unmark the same object again
-		 * without an exception being thrown. */
-		if (marked)
-		{
-			do
-			{
-				pdf_unmark_obj(ctx, node2);
-				if (node2 == marked)
-					break;
-				node2 = pdf_dict_get(ctx, node2, PDF_NAME(Parent));
-			}
-			while (node2);
-		}
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
-	}
-
-	return val;
+static pdf_obj *
+pdf_dict_getp_inheritable_imp(fz_context *ctx, pdf_obj *node, const char *path, int depth, pdf_cycle_list *cycle_up)
+{
+	pdf_cycle_list cycle;
+	pdf_obj *val = pdf_dict_getp(ctx, node, path);
+	if (val)
+		return val;
+	if (pdf_cycle(ctx, &cycle, cycle_up, node))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in tree (parents)");
+	if (depth > 100)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "too much recursion in tree (parents)");
+	node = pdf_dict_get(ctx, node, PDF_NAME(Parent));
+	if (node)
+		return pdf_dict_getp_inheritable_imp(ctx, node, path, depth + 1, &cycle);
+	return NULL;
 }
 
 pdf_obj *
 pdf_dict_getp_inheritable(fz_context *ctx, pdf_obj *node, const char *path)
 {
-	pdf_obj *node2 = node;
-	pdf_obj *val = NULL;
-	pdf_obj *marked = NULL;
-
-	fz_var(node);
-	fz_var(marked);
-	fz_try(ctx)
-	{
-		do
-		{
-			val = pdf_dict_getp(ctx, node, path);
-			if (val)
-				break;
-			if (pdf_mark_obj(ctx, node))
-				fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in tree (parents)");
-			marked = node;
-			node = pdf_dict_get(ctx, node, PDF_NAME(Parent));
-		}
-		while (node);
-	}
-	fz_always(ctx)
-	{
-		/* We assume that if we have marked an object, without an exception
-		 * being thrown, that we can always unmark the same object again
-		 * without an exception being thrown. */
-		if (marked)
-		{
-			do
-			{
-				pdf_unmark_obj(ctx, node2);
-				if (node2 == marked)
-					break;
-				node2 = pdf_dict_get(ctx, node2, PDF_NAME(Parent));
-			}
-			while (node2);
-		}
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
-	}
-
-	return val;
+	return pdf_dict_getp_inheritable_imp(ctx, node, path, 0, NULL);
 }
 
 void pdf_dict_put_bool(fz_context *ctx, pdf_obj *dict, pdf_obj *key, int x)
